@@ -43,7 +43,7 @@ namespace PerCederberg.Grammatica.Parser.RE {
      * maximum repetition count.
      *
      * @author   Per Cederberg, <per at percederberg dot net>
-     * @version  1.0
+     * @version  1.1
      */
     internal class RepeatElement : Element {
 
@@ -159,11 +159,15 @@ namespace PerCederberg.Grammatica.Parser.RE {
                                   int start, 
                                   int skip) {
 
+            if (skip == 0) {
+                matchStart = -1;
+                matches = null;
+            }
             switch (type) {
             case RepeatType.GREEDY:
                 return MatchGreedy(m, str, start, skip);
             case RepeatType.RELUCTANT:
-                return MatchReluctant(m, str, start, 1, skip, 0);
+                return MatchReluctant(m, str, start, skip);
             case RepeatType.POSSESSIVE:
                 if (skip == 0) {
                     return MatchPossessive(m, str, start, 0);
@@ -216,77 +220,40 @@ namespace PerCederberg.Grammatica.Parser.RE {
         }
 
         /**
-         * Returns the length of the minimum number of elements matching 
-         * the string starting at the specified position. The number of 
-         * matches to skip here and in the rest of the matches must also 
-         * be specified.
+         * Returns the length of the shortest possible matchine string
+         * starting at the specified position. The number of matches to 
+         * skip can also be specified.
          *
          * @param m              the matcher being used 
          * @param str            the string to match
          * @param start          the starting position
-         * @param count          the number of matched elements
-         * @param skipHere       the number of skips at this position
-         * @param skipNext       the number of skips at later positions
+         * @param skip           the number of matches to skip
          * 
-         * @return the length of the longest matching string, or
+         * @return the length of the shortest matching string, or
          *         -1 if no match was found
          */
         private int MatchReluctant(Matcher m, 
                                    string str, 
                                    int start,
-                                   int count,
-                                   int skipHere, 
-                                   int skipNext) {
-
-            int  length;
-            int  nextLength;
-
-            // Check for empty match
-            if (count == 1 && min == 0 && skipNext == 0) {
-                if (skipHere == 0) {
-                    return 0;
-                } else {
-                    skipHere--;
-                }
+                                   int skip) {
+    
+            // Find all matches
+            if (matchStart != start) {
+                matchStart = start;
+                matches = new BitArray(10);
+                FindMatches(m, str, start, 0, 0, 0);
             }
 
-            // Check element match
-            length = elem.Match(m, str, start, skipHere);
-
-            // Check for possible match here
-            if (length >= 0) {
-                if (min <= count) {
-                    if (skipNext == 0) {
-                        return length;
-                    } else {
-                        skipNext--;
+            // Find first non-skipped match
+            for (int i = 0; i < matches.Count; i++) {
+                if (matches[i]) {
+                    if (skip == 0) {
+                        return i;
                     }
-                }
-                if (length == 0) {
-                    return -1;
-                }
-                nextLength = MatchReluctant(m, 
-                                            str, 
-                                            start + length, 
-                                            count + 1, 
-                                            skipNext,
-                                            0);
-                if (nextLength >= 0) {
-                    return length + nextLength;
+                    skip--;
                 }
             }
-            
-            // Check for subsequent matches
-            if (skipHere == 0) {
-                return -1;
-            } else {
-                return MatchReluctant(m, 
-                                      str, 
-                                      start,
-                                      count, 
-                                      skipHere - 1, 
-                                      skipNext + 1);
-            }
+            return -1;
         }
 
         /**
