@@ -49,7 +49,7 @@ namespace PerCederberg.Grammatica.Parser {
      * cause infinite loop).
      *
      * @author   Per Cederberg, <per at percederberg dot net>
-     * @version  1.0
+     * @version  1.1
      */
     internal class LookAheadSet {
 
@@ -187,9 +187,6 @@ namespace PerCederberg.Grammatica.Parser {
          *  
          * @return true if the next tokens are in the set, or
          *         false otherwise
-         * 
-         * @throws ParseException if the next token couldn't be read 
-         *            or parsed correctly
          */
         public bool IsNext(Parser parser) {
             Sequence  seq;
@@ -212,9 +209,6 @@ namespace PerCederberg.Grammatica.Parser {
          *  
          * @return true if the next tokens are in the set, or
          *         false otherwise
-         * 
-         * @throws ParseException if the next token couldn't be read 
-         *            or parsed correctly
          */
         public bool IsNext(Parser parser, int length) {
             Sequence  seq;
@@ -228,6 +222,49 @@ namespace PerCederberg.Grammatica.Parser {
             return false;
         }
     
+        /**
+         * Checks if another look-ahead set has an overlapping token
+         * sequence. An overlapping token sequence is a token sequence
+         * that is identical to another sequence, but for the length.
+         * I.e. one of the two sequences may be longer than the other.
+         * 
+         * @param set            the look-ahead set to check
+         * 
+         * @return true if there is some token sequence that overlaps, or
+         *         false otherwise
+         */
+        public bool IsOverlap(LookAheadSet set) {
+            for (int i = 0; i < elements.Count; i++) {
+                if (set.IsOverlap((Sequence) elements[i])) {
+                    return true;
+                }
+            }
+            return false;
+        }
+    
+        /**
+         * Checks if a token sequence is overlapping. An overlapping token
+         * sequence is a token sequence that is identical to another 
+         * sequence, but for the length. I.e. one of the two sequences may
+         * be longer than the other. 
+         * 
+         * @param seq            the token sequence to check
+         * 
+         * @return true if there is some token sequence that overlaps, or
+         *         false otherwise
+         */
+        private bool IsOverlap(Sequence seq) {
+            Sequence  elem;
+
+            for (int i = 0; i < elements.Count; i++) {
+                elem = (Sequence) elements[i];
+                if (seq.StartsWith(elem) || elem.StartsWith(seq)) {
+                    return true;
+                }
+            }
+            return false;
+        }
+
         /**
          * Checks if the specified token sequence is present in the
          * set.
@@ -324,6 +361,27 @@ namespace PerCederberg.Grammatica.Parser {
             Add(new Sequence());
         }
 
+        /**
+         * Removes a token sequence from this set.
+         * 
+         * @param seq            the token sequence to remove
+         */
+        private void Remove(Sequence seq) {
+            elements.Remove(seq);
+        }
+
+        /**
+         * Removes all the token sequences from a specified set. Only 
+         * sequences already in this set will be removed.
+         * 
+         * @param set            the set to remove from
+         */
+        public void RemoveAll(LookAheadSet set) {
+            for (int i = 0; i < set.elements.Count; i++) {
+                Remove((Sequence) set.elements[i]);
+            }
+        }
+
         /**
          * Creates a new look-ahead set that is the result of reading
          * the specified token. The new look-ahead set will contain
@@ -417,6 +475,60 @@ namespace PerCederberg.Grammatica.Parser {
             return result;
         }
 
+        /**
+         * Creates a new look-ahead set with overlaps from another. All
+         * token sequences in this set that overlaps with the other set
+         * will be added to the new look-ahead set.
+         * 
+         * @param set            the look-ahead set to check with
+         * 
+         * @return a new look-ahead set containing the overlaps
+         */
+        public LookAheadSet CreateOverlaps(LookAheadSet set) {
+            LookAheadSet  result = new LookAheadSet(maxLength);
+            Sequence      seq;
+        
+            for (int i = 0; i < elements.Count; i++) {
+                seq = (Sequence) elements[i];
+                if (set.IsOverlap(seq)) {
+                    result.Add(seq);
+                }
+            }
+            return result;
+        }
+
+        /**
+         * Creates a new look-ahead set filter. The filter will contain
+         * all sequences from this set, possibly left trimmed by each one
+         * of the sequences in the specified set.
+         * 
+         * @param set            the look-ahead set to trim with
+         * 
+         * @return a new look-ahead set filter
+         */
+        public LookAheadSet CreateFilter(LookAheadSet set) {
+            LookAheadSet  result = new LookAheadSet(maxLength);
+            Sequence      first;
+            Sequence      second;
+        
+            // Handle special cases
+            if (this.Size() <= 0 || set.Size() <= 0) {
+                return this;
+            }
+
+            // Create combinations
+            for (int i = 0; i < elements.Count; i++) {
+                first = (Sequence) elements[i];
+                for (int j = 0; j < set.elements.Count; j++) {
+                    second = (Sequence) set.elements[j];
+                    if (first.StartsWith(second)) {
+                        result.Add(first.Subsequence(second.Length()));
+                    }
+                }
+            }
+            return result;
+        }
+
         /**
          * Creates a new identical look-ahead set, except for the
          * repeat flag being set in each token sequence.
@@ -444,18 +556,31 @@ namespace PerCederberg.Grammatica.Parser {
          * @return a string representation of this object
          */
         public override string ToString() {
+            return ToString(null);
+        }
+    
+        /**
+         * Returns a string representation of this object.
+         * 
+         * @param tokenizer      the tokenizer containing the tokens
+         * 
+         * @return a string representation of this object
+         */
+        public string ToString(Tokenizer tokenizer) {
             StringBuilder  buffer = new StringBuilder();
-            
+            Sequence       seq;
+
             buffer.Append("{");
             for (int i = 0; i < elements.Count; i++) {
+                seq = (Sequence) elements[i];
                 buffer.Append("\n  ");
-                buffer.Append(elements[i]);
+                buffer.Append(seq.ToString(tokenizer));
             }
             buffer.Append("\n}");
             return buffer.ToString();
         }
-    
-    
+
+
         /**
          * A token sequence. This class contains a list of token ids.
          * It is immutable after creation, meaning that no changes
@@ -597,6 +722,28 @@ namespace PerCederberg.Grammatica.Parser {
                 return true;
             }
 
+            /**
+             * Checks if this token sequence starts with the tokens from
+             * another sequence. If the other sequence is longer than this
+             * sequence, this method will always return false.
+             * 
+             * @param seq            the token sequence to check
+             * 
+             * @return true if this sequence starts with the other, or
+             *         false otherwise
+             */
+            public bool StartsWith(Sequence seq) {
+                if (Length() < seq.Length()) {
+                    return false;
+                }
+                for (int i = 0; i < seq.tokens.Count; i++) {
+                    if (!tokens[i].Equals(seq.tokens[i])) {
+                        return false;
+                    }
+                }
+                return true;
+            }
+
             /**
              * Checks if this token sequence is repetitive. A repetitive 
              * token sequence is one with the repeat flag set.
@@ -616,9 +763,6 @@ namespace PerCederberg.Grammatica.Parser {
              * 
              * @return true if the next tokens are in the sequence, or
              *         false otherwise
-             * 
-             * @throws ParseException if the next token couldn't be read 
-             *            or parsed correctly
              */
             public bool IsNext(Parser parser) {
                 Token   token;
@@ -643,9 +787,6 @@ namespace PerCederberg.Grammatica.Parser {
              * 
              * @return true if the next tokens are in the sequence, or
              *         false otherwise
-             * 
-             * @throws ParseException if the next token couldn't be read 
-             *            or parsed correctly
              */
             public bool IsNext(Parser parser, int length) {
                 Token  token;
@@ -670,11 +811,39 @@ namespace PerCederberg.Grammatica.Parser {
              * @return a string representation of this object
              */
             public override string ToString() {
-                if (repeat) {
-                    return tokens.ToString() + " *";
+                return ToString(null);
+            }
+        
+            /**
+             * Returns a string representation of this object.
+             *
+             * @param tokenizer      the tokenizer containing the tokens
+             *  
+             * @return a string representation of this object
+             */
+            public string ToString(Tokenizer tokenizer) {
+                StringBuilder  buffer = new StringBuilder();
+                string         str;
+                int            id;
+
+                if (tokenizer == null) {
+                    buffer.Append(tokens.ToString());
                 } else {
-                    return tokens.ToString();
+                    buffer.Append("[");
+                    for (int i = 0; i < tokens.Count; i++) {
+                        id = (int) tokens[i];
+                        str = tokenizer.GetPatternDescription(id);
+                        if (i > 0) {
+                            buffer.Append(" ");
+                        }
+                        buffer.Append(str);
+                    }
+                    buffer.Append("]");
                 }
+                if (repeat) {
+                    buffer.Append(" *");
+                }
+                return buffer.ToString();
             }
         
             /**
