@@ -41,7 +41,7 @@ namespace PerCederberg.Grammatica.Parser {
      * A parse exception.
      *
      * @author   Per Cederberg, <per at percederberg dot net>
-     * @version  1.0
+     * @version  1.1
      */
     public class ParseException : Exception {
 
@@ -225,7 +225,7 @@ namespace PerCederberg.Grammatica.Parser {
          * 
          * @return a default error message
          */
-        private string GetMessage() {
+        public string GetMessage() {
             StringBuilder  buffer = new StringBuilder();
 
             // Add error description
@@ -285,6 +285,10 @@ namespace PerCederberg.Grammatica.Parser {
                 break;
             default:
                 buffer.Append("internal error");
+                if (info != null) {
+                	buffer.Append(": ");
+                	buffer.Append(info);
+                }
                 break;
             }
 
