@@ -184,7 +184,7 @@ namespace PerCederberg.Grammatica.Parser {
                     return ParseAlternative(alt);
                 }
             }
-            if (defaultAlt == null) {
+            if (defaultAlt == null || !IsNext(defaultAlt)) {
                 ThrowParseException(FindUnion(pattern)); 
             }
             return ParseAlternative(defaultAlt);
