@@ -46,14 +46,22 @@ namespace PerCederberg.Grammatica.Parser {
      * don't match any of the token patterns, a parse exception is thrown. 
      *
      * @author   Per Cederberg, <per at percederberg dot net>
-     * @version  1.1
+     * @version  1.2
      */
     public class Tokenizer {
     
         /**
-         * The list of all token pattern matchers.
+         * The string token matcher. This token matcher is used for all
+         * string token patterns. This matcher implements a DFA to 
+         * provide maximum performance.
          */
-        private ArrayList matchers = new ArrayList();
+        private StringTokenMatcher stringMatcher = new StringTokenMatcher();
+
+        /**
+         * The list of all regular expression token matchers. These 
+         * matchers each test matches for a single regular expression.
+         */
+        private ArrayList regexpMatchers = new ArrayList();
 
         /**
          * The input stream to read from. When this is set to null, no
@@ -110,12 +118,17 @@ namespace PerCederberg.Grammatica.Parser {
          *         null if not present
          */
         public string GetPatternDescription(int id) {
-            TokenMatcher  m;
+            TokenPattern        pattern;
+            RegExpTokenMatcher  re;
         
-            for (int i = 0; i < matchers.Count; i++) {
-                m = (TokenMatcher) matchers[i];
-                if (m.GetPattern().GetId() == id) {
-                    return m.GetPattern().ToShortString();
+            pattern = stringMatcher.GetPattern(id);
+            if (pattern != null) {
+                return pattern.ToShortString();
+            }
+            for (int i = 0; i < regexpMatchers.Count; i++) {
+                re = (RegExpTokenMatcher) regexpMatchers[i];
+                if (re.GetPattern().GetId() == id) {
+                    return re.GetPattern().ToShortString();
                 }
             }
             return null;
@@ -154,11 +167,11 @@ namespace PerCederberg.Grammatica.Parser {
         public void AddPattern(TokenPattern pattern) {
             switch (pattern.GetPatternType()) {
             case TokenPattern.PatternType.STRING:
-                matchers.Add(new StringTokenMatcher(pattern));
+                stringMatcher.AddPattern(pattern);
                 break;
             case TokenPattern.PatternType.REGEXP:
                 try {
-                    matchers.Add(new RegExpTokenMatcher(pattern));
+                    regexpMatchers.Add(new RegExpTokenMatcher(pattern));
                 } catch (RegExpException e) {
                     throw new ParserCreationException(
                         ParserCreationException.ErrorType.INVALID_TOKEN,
@@ -223,10 +236,10 @@ namespace PerCederberg.Grammatica.Parser {
          *             parsed correctly
          */
         private Token NextToken() {
-            TokenMatcher   m;
-            Token          token;
-            string         str;
-           ParseException  e;
+            TokenMatcher    m;
+            Token           token;
+            string          str;
+            ParseException  e;
             
             // Find longest matching string 
             do {
@@ -240,9 +253,9 @@ namespace PerCederberg.Grammatica.Parser {
             // Return token results
             if (m != null) {
                 str = buffer.ToString();
-                str = str.Substring(position, m.Length());
-                token = new Token(m.GetPattern(), str, line, column); 
-                position += m.Length();
+                str = str.Substring(position, m.GetMatchedLength());
+                token = new Token(m.GetMatchedPattern(), str, line, column); 
+                position += m.GetMatchedLength();
                 line = token.GetEndLine();
                 column = token.GetEndColumn() + 1;
                 return token;
@@ -320,18 +333,30 @@ namespace PerCederberg.Grammatica.Parser {
          *         null if no match was found
          */
         private TokenMatcher FindMatch() {
-            TokenMatcher  bestMatch = null;
-            TokenMatcher  m;
-            string        str = buffer.ToString();
-            
-            for (int i = 0; i < matchers.Count; i++) {
-                m = (TokenMatcher) matchers[i];
-                if (m.MatchFrom(str, position)
-                 && (bestMatch == null || m.Length() > bestMatch.Length())) {
-                    
-                    bestMatch = m;
+            TokenMatcher        bestMatch = null;
+            int                 bestLength = 0;
+            RegExpTokenMatcher  re;
+            string              str = buffer.ToString();
+
+            // Check string matches
+            if (stringMatcher.MatchFrom(str, position)) {
+                bestMatch = stringMatcher;
+                bestLength = bestMatch.GetMatchedLength();
+            }
+            if (stringMatcher.HasReadEndOfString()) {
+                endOfBuffer = true;
+            }
+        
+            // Check regular expression matches
+            for (int i = 0; i < regexpMatchers.Count; i++) {
+                re = (RegExpTokenMatcher) regexpMatchers[i];
+                if (re.MatchFrom(str, position)
+                 && re.GetMatchedLength() > bestLength) {
+
+                    bestMatch = re;
+                    bestLength = bestMatch.GetMatchedLength();
                 }
-                if (m.HasReadEndOfString()) {
+                if (re.HasReadEndOfString()) {
                     endOfBuffer = true;
                 }
             }
@@ -347,12 +372,10 @@ namespace PerCederberg.Grammatica.Parser {
          */
         public override string ToString() {
             StringBuilder  buffer = new StringBuilder();
-            TokenMatcher   m;
 
-            for (int i = 0; i < matchers.Count; i++) {
-                m = (TokenMatcher) matchers[i];
-                buffer.Append(m);
-                buffer.Append('\n');
+            buffer.Append(stringMatcher);
+            for (int i = 0; i < regexpMatchers.Count; i++) {
+                buffer.Append(regexpMatchers[i]);
             }
             return buffer.ToString();
         }
@@ -367,20 +390,68 @@ namespace PerCederberg.Grammatica.Parser {
      */
     internal abstract class TokenMatcher {
 
+        /**
+         * Returns the latest matched token pattern.
+         * 
+         * @return the latest matched token pattern, or
+         *         null if no match found
+         */
+        public abstract TokenPattern GetMatchedPattern();
+        
+        /**
+         * Returns the length of the latest match.
+         * 
+         * @return the length of the latest match, or 
+         *         zero (0) if no match found
+         */
+        public abstract int GetMatchedLength();
+
+        /**
+         * Checks if the end of string was encountered during the last
+         * match. 
+         * 
+         * @return true if the end of string was reached, or
+         *         false otherwise
+         */
+        public abstract bool HasReadEndOfString();
+    }
+
+
+    /**
+     * A regular expression token pattern matcher. This class is used
+     * to match a single regular expression with the tokenizer 
+     * buffer. This class also maintains the state of the last match. 
+     */
+    internal class RegExpTokenMatcher : TokenMatcher {
+
         /**
          * The token pattern to match with.
          */
         private TokenPattern pattern;
         
         /**
-         * Creates a new token pattern matcher.
-         * 
-         * @param pattern        the pattern to match with
+         * The regular expression to use.
          */
-        public TokenMatcher(TokenPattern pattern) {
+        private RegExp regExp;
+
+        /**
+         * The regular expression matcher to use.
+         */
+        private Matcher matcher = null;
+
+        /**
+         * Creates a new regular expression token matcher.
+         * 
+         * @param pattern        the pattern to match
+         * 
+         * @throws RegExpException if the regular expression couldn't
+         *             be created properly
+         */
+        public RegExpTokenMatcher(TokenPattern pattern) {
             this.pattern = pattern;
+            this.regExp = new RegExp(pattern.GetPattern());
         }
-        
+
         /**
          * Returns the token pattern.
          * 
@@ -396,191 +467,7 @@ namespace PerCederberg.Grammatica.Parser {
          * @return the start position of the last match, or
          *         zero (0) if none found
          */
-        public abstract int Start();
-
-        /**
-         * Returns the length of the latest match.
-         * 
-         * @return the length of the last match, or
-         *         zero (0) if none found
-         */        
-        public abstract int Length();
-        
-        /**
-         * Checks if the end of string was encountered during the last
-         * match. 
-         * 
-         * @return true if the end of string was reached, or
-         *         false otherwise
-         */
-        public abstract bool HasReadEndOfString(); 
-
-        /**
-         * Checks if the token pattern matches a string from a
-         * specified position. This method will also reset all flags
-         * in this matcher.
-         * 
-         * @param str            the string to match
-         * @param pos            the starting position
-         * 
-         * @return true if a match was found, or
-         *         false otherwise
-         */
-        public abstract bool MatchFrom(string str, int pos);
-
-        /**
-         * Returns a string representation of this token matcher.
-         * 
-         * @return a detailed string representation of this matcher
-         */
-        public override string ToString() {
-            return pattern.ToString() + "\n";
-        }
-    }
-    
-
-    /**
-     * A string token pattern matcher. This class is used to check 
-     * for matches with the tokenizer buffer, and maintain the state 
-     * of the last match. 
-     */
-    internal class StringTokenMatcher : TokenMatcher {
-
-        /**
-         * The string image to search for.
-         */
-        private string image;
-        
-        /**
-         * The start position of the last match. This will be a 
-         * negative value if no match was found.
-         */
-        private int start;
-
-        /**
-         * The end of string read flag.
-         */
-        private bool endOfString;
-
-        /**
-         * Creates a new string token matcher.
-         * 
-         * @param pattern        the token pattern to use
-         */
-        public StringTokenMatcher(TokenPattern pattern) 
-            : base(pattern) {
-
-            image = pattern.GetPattern();
-            Reset();
-        }
-
-        /**
-         * Resets the matcher state. This will clear the results of 
-         * the last match. 
-         */
-        public void Reset() {
-            start = -1;
-            endOfString = false;
-        }
-
-        /**
-         * Returns the start position of the latest match.
-         * 
-         * @return the start position of the last match, or
-         *         zero (0) if none found
-         */
-        public override int Start() {
-            return (start < 0) ? 0 : start;
-        }
-        
-        /**
-         * Returns the length of the latest match.
-         * 
-         * @return the length of the last match, or
-         *         zero (0) if none found
-         */        
-        public override int Length() {
-            return (start < 0) ? 0 : image.Length;
-        }
-        
-        /**
-         * Checks if the end of string was encountered during the last
-         * match. 
-         * 
-         * @return true if the end of string was reached, or
-         *         false otherwise
-         */
-        public override bool HasReadEndOfString() {
-            return endOfString; 
-        }
-
-        /**
-         * Checks if the token pattern matches a string from a
-         * specified position. This method will also reset all flags
-         * in this matcher.
-         * 
-         * @param str            the string to match
-         * @param pos            the starting position
-         * 
-         * @return true if a match was found, or
-         *         false otherwise
-         */
-        public override bool MatchFrom(string str, int pos) {
-            Reset();
-            if (pos + image.Length > str.Length) {
-                endOfString = true;
-                return false;
-            }
-            for (int i = 0; i < image.Length; i++) {
-                if (image[i] != str[pos + i]) {
-                    return false;
-                }
-            }
-            start = pos;
-            return true;
-        }
-    }
-    
-
-    /**
-     * A regular expression token pattern matcher. This class is used
-     * to check for matches with the tokenizer buffer, and maintain 
-     * the state of the last match. 
-     */
-    internal class RegExpTokenMatcher : TokenMatcher {
-
-        /**
-         * The regular expression to use.
-         */
-        private RegExp regExp;
-
-        /**
-         * The last regular expression matcher used. This variable is
-         * null when this matcher has been reset.
-         */
-        private Matcher matcher;
-
-        /**
-         * Creates a new regular expression token matcher.
-         * 
-         * @param pattern        the pattern to match
-         * 
-         * @throws RegExpException if the regular expression couldn't
-         *             be created properly
-         */
-        public RegExpTokenMatcher(TokenPattern pattern)
-            : base(pattern) {
-
-            regExp = new RegExp(pattern.GetPattern());
-        }
-
-        /**
-         * Returns the start position of the latest match.
-         * 
-         * @return the start position of the last match, or
-         *         zero (0) if none found
-         */
-        public override int Start() {
+        public int Start() {
             if (matcher == null || matcher.Length() <= 0) {
                 return 0;
             } else {
@@ -588,13 +475,27 @@ namespace PerCederberg.Grammatica.Parser {
             }
         }
         
+        /**
+         * Returns the latest matched token pattern.
+         * 
+         * @return the latest matched token pattern, or
+         *         null if no match found
+         */
+        public override TokenPattern GetMatchedPattern() {
+            if (matcher == null || matcher.Length() <= 0) {
+                return null;
+            } else {
+                return pattern;
+            }
+        }
+        
         /**
          * Returns the length of the latest match.
          * 
-         * @return the length of the last match, or
-         *         zero (0) if none found
-         */        
-        public override int Length() {
+         * @return the length of the latest match, or 
+         *         zero (0) if no match found
+         */
+        public override int GetMatchedLength() {
             return (matcher == null) ? 0 : matcher.Length();
         }
         
@@ -610,9 +511,9 @@ namespace PerCederberg.Grammatica.Parser {
         }
 
         /**
-         * Checks if the token pattern matches a string from a
-         * specified position. This method will also reset all flags
-         * in this matcher.
+         * Checks if the token pattern matches the tokenizer buffer  
+         * from the specified position. This method will also reset 
+         * all flags in this matcher.
          * 
          * @param str            the string to match
          * @param pos            the starting position
@@ -620,18 +521,326 @@ namespace PerCederberg.Grammatica.Parser {
          * @return true if a match was found, or
          *         false otherwise
          */
-        public override bool MatchFrom(string str, int pos) {
+        public bool MatchFrom(string str, int pos) {
             matcher = regExp.Matcher(str);
             return matcher.MatchFrom(pos);
         }
-
+        
         /**
          * Returns a string representation of this token matcher.
          * 
          * @return a detailed string representation of this matcher
          */
         public override string ToString() {
-            return base.ToString() + regExp.ToString();
+            return pattern.ToString() + "\n" + 
+                   regExp.ToString() + "\n";
+        }
+    }
+
+
+    /**
+     * A string token pattern matcher. This class is used to match a 
+     * set of strings with the tokenizer buffer. This class 
+     * internally uses a DFA for maximum performance. It also 
+     * maintains the state of the last match. 
+     */
+    internal class StringTokenMatcher : TokenMatcher {
+
+        /**
+         * The list of string token patterns. 
+         */
+        private ArrayList patterns = new ArrayList();
+
+        /**
+         * The finite automaton to use for matching.
+         */
+        private Automaton start = new Automaton();
+
+        /**
+         * The last token pattern match found.
+         */
+        private TokenPattern match = null;
+
+        /**
+         * The end of string read flag.
+         */
+        private bool endOfString = false;
+
+        /**
+         * Creates a new string token matcher.
+         */
+        public StringTokenMatcher() {
+        }
+
+        /**
+         * Resets the matcher state. This will clear the results of 
+         * the last match. 
+         */
+        public void Reset() {
+            match = null;
+            endOfString = false;
+        }
+
+        /**
+         * Returns the latest matched token pattern.
+         * 
+         * @return the latest matched token pattern, or
+         *         null if no match found
+         */
+        public override TokenPattern GetMatchedPattern() {
+            return match;
+        }
+        
+        /**
+         * Returns the length of the latest match.
+         * 
+         * @return the length of the latest match, or 
+         *         zero (0) if no match found
+         */
+        public override int GetMatchedLength() {
+            if (match == null) {
+                return 0;
+            } else {
+                return match.GetPattern().Length;
+            }
+        }
+
+        /**
+         * Checks if the end of string was encountered during the last
+         * match. 
+         * 
+         * @return true if the end of string was reached, or
+         *         false otherwise
+         */
+        public override bool HasReadEndOfString() {
+            return endOfString; 
+        }
+
+        /**
+         * Sets the end of string encountered flag.
+         */
+        public void SetReadEndOfString() {
+            endOfString = true;
+        }
+
+        /**
+         * Returns the token pattern with the specified id. Only 
+         * token patterns handled by this matcher can be returned. 
+         * 
+         * @param id         the token pattern id
+         * 
+         * @return the token pattern found, or
+         *         null if not found
+         */
+        public TokenPattern GetPattern(int id) {
+            TokenPattern  pattern;
+
+            for (int i = 0; i < patterns.Count; i++) {
+                pattern = (TokenPattern) patterns[i];
+                if (pattern.GetId() == id) {
+                    return pattern;
+                }
+            }
+            return null;
+        }
+
+        /**
+         * Adds a string token pattern to this matcher. 
+         * 
+         * @param pattern        the pattern to add
+         */
+        public void AddPattern(TokenPattern pattern) {
+            patterns.Add(pattern);
+            start.AddMatch(pattern.GetPattern(), pattern);
+        }
+
+        /**
+         * Checks if the token pattern matches the tokenizer buffer  
+         * from the specified position. This method will also reset 
+         * all flags in this matcher.
+         * 
+         * @param str            the string to match
+         * @param pos            the starting position
+         * 
+         * @return true if a match was found, or
+         *         false otherwise
+         */
+        public bool MatchFrom(string str, int pos) {
+            Reset();
+            match = (TokenPattern) start.MatchFrom(this, str, pos);
+            return match != null;
+        }
+        
+        /**
+         * Returns a string representation of this matcher. This will
+         * contain all the token patterns.
+         * 
+         * @return a detailed string representation of this matcher 
+         */
+        public override string ToString() {
+            StringBuilder  buffer = new StringBuilder();
+            
+            for (int i = 0; i < patterns.Count; i++) {
+                buffer.Append(patterns[i]);
+                buffer.Append("\n\n");
+            }
+            return buffer.ToString();
+        }
+    }
+    
+
+    /**
+     * A deterministic finite automaton. This is a simple automaton
+     * for character sequences. It cannot handle character set state
+     * transitions, but only supports single character transitions.
+     */
+    internal class Automaton {
+
+        /**
+         * The state value.
+         */
+        private object value = null;
+        
+        /**
+         * The automaton state transition tree. Each transition from
+         * this state to another state is added to this tree with the
+         * corresponding character. 
+         */
+        private AutomatonTree tree = new AutomatonTree();
+        
+        /**
+         * Creates a new empty automaton. 
+         */
+        public Automaton() {
+        }
+
+        /**
+         * Adds a string match to this automaton. New states and 
+         * transitions will be added to extend this automaton to 
+         * support the specified string.
+         * 
+         * @param str            the string to match
+         * @param value          the match value
+         */
+        public void AddMatch(string str, object value) {
+            Automaton  state;
+
+            if (str.Equals("")) {
+                this.value = value;
+            } else {
+                state = tree.Find(str[0]);
+                if (state == null) {
+                    state = new Automaton();
+                    state.AddMatch(str.Substring(1), value);
+                    tree.Add(str[0], state);
+                } else {
+                    state.AddMatch(str.Substring(1), value);
+                }
+            }
+        }
+
+        /**
+         * Checks if the automaton matches the tokenizer buffer from 
+         * the specified position. This method will set the end of
+         * buffer flag in the specified token matcher if the end of
+         * buffer is reached.
+         *
+         * @param m              the string token matcher 
+         * @param str            the string to match
+         * @param pos            the starting position
+         * 
+         * @return the match value, or
+         *         null if no match is found
+         */
+        public object MatchFrom(StringTokenMatcher m, string str, int pos) {
+            object     result = null;
+            Automaton  state;
+
+            if (pos >= str.Length) {
+                m.SetReadEndOfString();
+            } else if (tree != null) {
+                state = tree.Find(str[pos]);
+                if (state != null) {
+                    result = state.MatchFrom(m, str, pos + 1);
+                }
+            }
+            return (result == null) ? value : result;
+        }
+    }
+    
+    
+    /**
+     * An automaton state transition tree. This class contains a 
+     * binary search tree for the automaton transitions from one 
+     * state to another. All transitions are linked to a single 
+     * character.
+     */
+    internal class AutomatonTree {
+
+        /**
+         * The transition character. If this value is set to the zero 
+         * ('\0') character, this tree is empty.
+         */
+        private char value = '\0';
+        
+        /**
+         * The transition state.
+         */
+        private Automaton state = null;
+        
+        /**
+         * The left subtree.
+         */
+        private AutomatonTree left = null;
+        
+        /**
+         * The right subtree.
+         */
+        private AutomatonTree right = null;
+        
+        /**
+         * Creates a new empty automaton transition tree.
+         */
+        public AutomatonTree() {
+        }
+        
+        /**
+         * Finds an automaton state from the specified transition 
+         * character. This method searches this transition tree for
+         * a matching transition.
+         * 
+         * @param c              the character to search for
+         * 
+         * @return the automaton state found, or
+         *         null if no transition exists
+         */
+        public Automaton Find(char c) {
+            if (value == '\0' || value == c) {
+                return state;
+            } else if (value > c) {
+                return left.Find(c); 
+            } else {
+                return right.Find(c);
+            }
+        }
+        
+        /**
+         * Adds a transition to this tree. 
+         * 
+         * @param c              the character to transition for
+         * @param state          the state to transition to
+         */
+        public void Add(char c, Automaton state) {
+            if (value == '\0') {
+                this.value = c;
+                this.state = state;
+                this.left = new AutomatonTree();
+                this.right = new AutomatonTree();
+            } else if (value > c) {
+                left.Add(c, state);
+            } else {
+                right.Add(c, state);
+            }
         }
     }
 }
