@@ -38,9 +38,12 @@ import java.io.File;
 import java.io.FileNotFoundException;
 import java.io.FileReader;
 import java.io.IOException;
+import java.io.PrintWriter;
 
 import net.percederberg.grammatica.output.CSharpParserGenerator;
 import net.percederberg.grammatica.output.JavaParserGenerator;
+import net.percederberg.grammatica.parser.Analyzer;
+import net.percederberg.grammatica.parser.Node;
 import net.percederberg.grammatica.parser.ParseException;
 import net.percederberg.grammatica.parser.Parser;
 import net.percederberg.grammatica.parser.ParserCreationException;
@@ -54,7 +57,7 @@ import net.percederberg.grammatica.parser.Tokenizer;
  * for information on usage and command-line parameters.
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.0
+ * @version  1.2
  */
 public class Grammatica extends Object {
 
@@ -78,32 +81,39 @@ public class Grammatica extends Object {
         "  --parse <file>\n" +
         "      Debugs the grammar by using it to parse the specified\n" +
         "      file. No code has to be generated for this.\n" +
+        "  --profile <file>\n" +
+        "      Profiles the grammar by using it to parse the specified\n" +
+        "      file and printing a statistic summary.\n" +
         "  --csoutput <dir>\n" +
         "      Creates a C# parser for the grammar (in source code).\n" +
-        "      The specified directory with be used when generating\n" +
-        "      source code files.\n" +
+        "      The specified directory will be used as output directory\n" +
+        "      for the source code files.\n" +
         "  --javaoutput <dir>\n" +
         "      Creates a Java parser for the grammar (in source code).\n" +
-        "      The specified directory with be used as base directory\n" +
-        "      when generating source code files.\n" +
+        "      The specified directory will be used as the base output\n" +
+        "      directory for the source code files.\n" +
         "\n" +
         "C# Output Options:\n" +
         "  --csnamespace <package>\n" +
-        "      Sets the C# namespace to use in generated source code files.\n" +
+        "      Sets the C# namespace to use in generated source code\n" +
+        "      files. By default no namespace declaration is included.\n" +
         "  --csclassname <name>\n" +
         "      Sets the C# class name prefix to use in generated source\n" +
-        "      code files.\n" +
+        "      code files. By default the grammar file name is used.\n" +
         "  --cspublic\n" +
-        "      Sets public access for all C# types generated.\n" +
+        "      Sets public access for all C# types generated. By default\n" +
+        "      type access is internal.\n" +
         "\n" +
         "Java Output Options:\n" +
         "  --javapackage <package>\n" +
-        "      Sets the Java package to use in generated source code files.\n" +
+        "      Sets the Java package to use in generated source code\n" +
+        "      files. By default no package declaration is included.\n" +
         "  --javaclassname <name>\n" +
         "      Sets the Java class name prefix to use in generated source\n" +
-        "      code files.\n" +
+        "      code files. By default the grammar file name is used.\n" +
         "  --javapublic\n" +
-        "      Sets public access for all Java types.";
+        "      Sets public access for all Java types. By default type\n" +
+        "      access is package local.";
 
     /**
      * The internal error message.
@@ -162,6 +172,8 @@ public class Grammatica extends Object {
                 tokenize(grammar, new File(args[2]));
             } else if (args[1].equals("--parse")) {
                 parse(grammar, new File(args[2]));
+            } else if (args[1].equals("--profile")) {
+                profile(grammar, new File(args[2]));
             } else if (args[1].equals("--javaoutput")) {
                 writeJavaCode(args, grammar);
             } else if (args[1].equals("--csoutput")) {
@@ -458,14 +470,16 @@ public class Grammatica extends Object {
      */
     private static void parse(Grammar grammar, File file) {
         Tokenizer  tokenizer;
+        Analyzer   analyzer;
         Parser     parser;
         Token      token;
         
         try {
             tokenizer = grammar.createTokenizer(new FileReader(file));
-            parser = grammar.createParser(tokenizer);
+            analyzer = new TreePrinter(new PrintWriter(System.out));
+            parser = grammar.createParser(tokenizer, analyzer);
             System.out.println("Parse tree from " + file + ":");
-            parser.parse().printTo(System.out);
+            parser.parse();
         } catch (FileNotFoundException e) {
             printError(file.toString(), e);
             System.exit(1);
@@ -481,6 +495,75 @@ public class Grammatica extends Object {
         }
     }
 
+    /**
+     * Parses the specified file with the grammar and prints 
+     * profiling information.
+     * 
+     * @param grammar        the grammar to use
+     * @param file           the file to parse
+     */
+    private static void profile(Grammar grammar, File file) {
+        Tokenizer  tokenizer;
+        Parser     parser;
+        Node       node;
+        long       time;
+        int        counter;
+        
+        // Profile tokenizer
+        try {
+            tokenizer = grammar.createTokenizer(new FileReader(file));
+            System.out.println("Tokenizing " + file);
+            time = System.currentTimeMillis();
+            counter = 0;
+            while (tokenizer.next() != null) {
+                counter++;
+            }
+            time = System.currentTimeMillis() - time;
+            System.out.println("  Time elapsed:  " + time + " millisec");
+            System.out.println("  Tokens found:  " + counter);
+            System.out.println("  Average speed: " + (counter / time) +
+                               " tokens/millisec");
+            System.out.println();
+        } catch (FileNotFoundException e) {
+            printError(file.toString(), e);
+            System.exit(1);
+        } catch (GrammarException e) {
+            printInternalError(e);
+            System.exit(2);
+        } catch (ParseException e) {
+            printError(file.toString(), e);
+            System.exit(1);
+        }
+
+        // Profile parser
+        try {
+            tokenizer = grammar.createTokenizer(new FileReader(file));
+            parser = grammar.createParser(tokenizer);
+            System.out.println("Parsing " + file);
+            time = System.currentTimeMillis();
+            node = parser.parse();
+            time = System.currentTimeMillis() - time;
+            counter = 1 + node.getDescendantCount();
+            System.out.println("  Time elapsed:  " + time + " millisec");
+            System.out.println("  Nodes found:   " + counter);
+            System.out.println("  Average speed: " + (counter / time) +
+                               " nodes/millisec");
+            System.out.println();
+        } catch (FileNotFoundException e) {
+            printError(file.toString(), e);
+            System.exit(1);
+        } catch (GrammarException e) {
+            printInternalError(e);
+            System.exit(2);
+        } catch (ParserCreationException e) {
+            printInternalError(e);
+            System.exit(2);
+        } catch (ParserLogException e) {
+            printError(file.toString(), e);
+            System.exit(1);
+        }
+    }
+    
     /**
      * Parses the command-line arguments and generates the Java source 
      * code for a parser. 
@@ -554,4 +637,62 @@ public class Grammatica extends Object {
             System.exit(1);
         }
     }
+    
+    
+    /**
+     * A parse tree printer. This class prints the parse tree while
+     * it is being parsed. 
+     */
+    private static class TreePrinter extends Analyzer {
+
+        /**
+         * The current indentation level.
+         */
+        private int indentation = 0; 
+
+        /**
+         * The output stream to use.
+         */
+        private PrintWriter output;
+
+        /**
+         * Creates a new parse tree printer.
+         * 
+         * @param output         the output stream to use
+         */
+        public TreePrinter(PrintWriter output) {
+            this.output = output;
+        }
+
+        /**
+         * Called when entering a parse tree node. By default this method
+         * does nothing. A subclass can override this method to handle 
+         * each node separately.  
+         * 
+         * @param node           the node being entered
+         */
+        protected void enter(Node node) {
+            for (int i = 0; i < indentation; i++) {
+                output.print("  ");
+            }
+            output.println(node.toString());
+            output.flush();
+            indentation++;
+        }
+    
+        /**
+         * Called when exiting a parse tree node. By default this method
+         * returns the node. A subclass can override this method to handle 
+         * each node separately. If no parse tree should be created, this 
+         * method should return null.
+         * 
+         * @param node           the node being exited
+         * 
+         * @return the node to add to the parse tree
+         */
+        protected Node exit(Node node) {
+            indentation--;
+            return null;
+        }
+    }
 }
