@@ -36,17 +36,21 @@ package net.percederberg.grammatica.parser;
 import java.util.ArrayList;
 import java.util.Collection;
 import java.util.HashMap;
-import java.util.Iterator;
 
 /**
  * A base parser class. This class provides the standard parser 
  * interface, as well as token handling.
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.0
+ * @version  1.1
  */
 public abstract class Parser {
 
+    /**
+     * The parser initialization flag.
+     */
+    private boolean initialized = false;
+
     /**
      * The tokenizer to use.
      */
@@ -58,22 +62,38 @@ public abstract class Parser {
     private Analyzer analyzer;
 
     /**
-     * The start production pattern id. This will be set to the first 
-     * production pattern added.
+     * The list of production patterns. 
      */
-    private int start = 0;
+    private ArrayList patterns = new ArrayList();
 
     /**
-     * The map with all production patterns. The production patterns 
-     * are indexed by their id.
+     * The map with production patterns and their id:s. This map 
+     * contains the production patterns indexed by their id:s.
      */
-    private HashMap patterns = new HashMap();
+    private HashMap patternIds = new HashMap();
 
     /**
      * The list of buffered tokens. This list will contain tokens that
      * have been read from the tokenizer, but not yet consumed.
      */
     private ArrayList tokens = new ArrayList();
+    
+    /**
+     * The error log. All parse errors will be added to this log as
+     * the parser attempts to recover from the error. If the error
+     * count is higher than zero (0), this log will be thrown as the
+     * result from the parse() method. 
+     */
+    private ParserLogException errorLog = new ParserLogException();
+
+    /**
+     * The error recovery counter. This counter is initially set to a
+     * negative value to indicate that no error requiring recovery 
+     * has been encountered. When a parse error is found, the counter
+     * is set to three (3), and is then decreased by one for each 
+     * correctly read token until it reaches zero (0).  
+     */
+    private int errorRecovery = -1;
 
     /**
      * Creates a new parser.
@@ -95,6 +115,17 @@ public abstract class Parser {
         this.analyzer = analyzer;
     }
 
+    /**
+     * Sets the parser initialized flag. Normally this flag is set by
+     * the prepare() method, but this method allows further 
+     * modifications to it.
+     * 
+     * @param initialized    the new initialized flag
+     */
+    void setInitialized(boolean initialized) {
+        this.initialized = initialized;
+    }
+
     /**
      * Adds a new production pattern to the parser. The first pattern 
      * added is assumed to be the starting point in the grammar. The 
@@ -117,17 +148,16 @@ public abstract class Parser {
                 "no production alternatives are present (must have at " +
                 "least one)");
         }
-        if (patterns.size() == 0) {
-            start = pattern.getId();
-        }
-        if (patterns.containsKey(id)) {
+        if (patternIds.containsKey(id)) {
             throw new ParserCreationException(
                 ParserCreationException.INVALID_PRODUCTION_ERROR,
                 pattern.getName(),
                 "another pattern with the same id (" + id + 
                 ") has already been added");
         }
-        patterns.put(id, pattern);
+        patterns.add(pattern);
+        patternIds.put(id, pattern);
+        setInitialized(false);
     }
 
     /**
@@ -139,17 +169,15 @@ public abstract class Parser {
      *             initialized correctly 
      */
     public void prepare() throws ParserCreationException {
-        Iterator  iter;
-
-        if (patterns.size() == 0) {
+        if (patterns.size() <= 0) {
             throw new ParserCreationException(
                 ParserCreationException.INVALID_PARSER_ERROR,
                 "no production patterns have been added");
         }
-        iter = patterns.values().iterator();
-        while (iter.hasNext()) {
-            checkPattern((ProductionPattern) iter.next());
+        for (int i = 0; i < patterns.size(); i++) {
+            checkPattern((ProductionPattern) patterns.get(i));
         }
+        setInitialized(true);
     }
 
     /**
@@ -214,19 +242,71 @@ public abstract class Parser {
 
     /**
      * Parses the token stream and returns a parse tree. This method
-     * will call prepare() if not previously called.
+     * will call prepare() if not previously called. In case of a 
+     * parse error, the parser will attempt to recover and throw all
+     * the errors found in a parser log exception in the end.
      * 
      * @return the parse tree
      * 
      * @throws ParserCreationException if the parser couldn't be
      *             initialized correctly
-     * @throws ParseException if the input couldn't be parsed 
+     * @throws ParserLogException if the input couldn't be parsed 
      *             correctly
      * 
      * @see #prepare
      */
-    public abstract Node parse() 
-        throws ParserCreationException, ParseException;
+    public Node parse() throws ParserCreationException, ParserLogException {
+        Node  root = null;
+
+        // Initialize parser
+        if (!initialized) {
+            prepare();
+        }
+
+        // Parse input
+        try {
+            root = parseStart();
+        } catch (ParseException e) {
+            addError(e, true);
+        }
+        
+        // Check for errors
+        if (errorLog.getErrorCount() > 0) {
+            throw errorLog;
+        }
+
+        return root;
+    }
+
+    /**
+     * Parses the token stream and returns a parse tree.
+     * 
+     * @return the parse tree
+     * 
+     * @throws ParseException if the input couldn't be parsed 
+     *             correctly
+     */
+    protected abstract Node parseStart() throws ParseException;
+
+    /**
+     * Adds an error to the error log. If the parser is in error 
+     * recovery mode, the error will not be added to the log. If the
+     * recovery flag is set, this method will set the error recovery 
+     * counter thus enter error recovery mode. Only lexical or 
+     * syntactical errors require recovery, so this flag shouldn't be
+     * set otherwise.
+     * 
+     * @param e              the error to add
+     * @param recovery       the recover flag 
+     */
+    void addError(ParseException e, boolean recovery) {
+        if (errorRecovery <= 0) {
+            errorLog.addError(e);
+        }
+        if (recovery) {
+            errorRecovery = 3;
+        }
+    }
 
     /**
      * Returns the production pattern with the specified id.
@@ -239,7 +319,7 @@ public abstract class Parser {
     ProductionPattern getPattern(int id) {
         Integer  value = new Integer(id);
         
-        return (ProductionPattern) patterns.get(value);
+        return (ProductionPattern) patternIds.get(value);
     }
 
     /**
@@ -249,75 +329,86 @@ public abstract class Parser {
      *         null if no patterns have been added
      */
     ProductionPattern getStartPattern() {
-        return getPattern(start);
+        if (patterns.size() <= 0) {
+            return null;
+        } else {
+            return (ProductionPattern) patterns.get(0);
+        }
     }
 
     /**
-     * Returns the unordered set of production patterns.
+     * Returns the ordered set of production patterns.
      * 
-     * @return the unordered set of production patterns
+     * @return the ordered set of production patterns
      */
     Collection getPatterns() {
-        return patterns.values();
+        return patterns;
     }
 
     /**
      * Handles the parser entering a production. This method calls the
-     * appropriate analyzer callback if the node is not hidden.
+     * appropriate analyzer callback if the node is not hidden. Note
+     * that this method will not call any callback if an error 
+     * requiring recovery has ocurred.
      * 
      * @param node           the parse tree node
-     * 
-     * @throws ParseException if the node couldn't be handled 
-     *             correctly
      */
-    void enterNode(Node node) throws ParseException {
-        if (!node.isHidden()) {
-            analyzer.enter(node);
+    void enterNode(Node node) {
+        if (!node.isHidden() && errorRecovery < 0) {
+            try {
+                analyzer.enter(node);
+            } catch (ParseException e) {
+                addError(e, false);
+            }
         }
     }
     
     /**
      * Handles the parser leaving a production. This method calls the
      * appropriate analyzer callback if the node is not hidden, and 
-     * returns the result.
+     * returns the result. Note that this method will not call any 
+     * callback if an error requiring recovery has ocurred.
      * 
      * @param node           the parse tree node
      * 
      * @return the parse tree node, or
      *         null if no parse tree should be created
-     * 
-     * @throws ParseException if the node couldn't be handled 
-     *             correctly
      */
-    Node exitNode(Node node) throws ParseException {
-        if (node.isHidden()) {
-            return node;
-        } else {
-            return analyzer.exit(node);
+    Node exitNode(Node node) {
+        if (!node.isHidden() && errorRecovery < 0) {
+            try {
+                return analyzer.exit(node);
+            } catch (ParseException e) {
+                addError(e, false);
+            }
         }
+        return node;
     }
 
     /**
      * Handles the parser adding a child node to a production. This 
-     * method calls the appropriate analyzer callback.
+     * method calls the appropriate analyzer callback. Note that this 
+     * method will not call any callback if an error requiring 
+     * recovery has ocurred.
      * 
      * @param node           the parent parse tree node
      * @param child          the child parse tree node, or null
-     * 
-     * @throws ParseException if the node couldn't be handled 
-     *             correctly
      */
-    void addNode(Production node, Node child) 
-        throws ParseException {
-
-        if (node.isHidden()) {
+    void addNode(Production node, Node child) {
+        if (errorRecovery >= 0) {
+            // Do nothing
+        } else if (node.isHidden()) {
             node.addChild(child);
         } else if (child != null && child.isHidden()) {
             for (int i = 0; i < child.getChildCount(); i++) {
                 addNode(node, child.getChildAt(i));
             }
         } else {
-            analyzer.child(node, child);
+            try {
+                analyzer.child(node, child);
+            } catch (ParseException e) {
+                addError(e, false);
+            }
         }
     }
 
@@ -363,6 +454,9 @@ public abstract class Parser {
         ArrayList  list;
         
         if (token.getId() == id) {
+            if (errorRecovery > 0) {
+                errorRecovery--;
+            }
             return token;
         } else {
             list = new ArrayList(1);
@@ -385,24 +479,143 @@ public abstract class Parser {
      * 
      * @return the token in the queue, or
      *         null if no more tokens in the queue
-     * 
-     * @throws ParseException if the input stream couldn't be read or
-     *             parsed correctly
      */
-    Token peekToken(int steps) throws ParseException {
+    Token peekToken(int steps) {
         Token  token;
 
         while (steps >= tokens.size()) {
-            token = tokenizer.next();
-            if (token == null) {
-                return null;
-            } else {
-                tokens.add(token);
+            try {
+                token = tokenizer.next();
+                if (token == null) {
+                    return null;
+                } else {
+                    tokens.add(token);
+                }
+            } catch (ParseException e) {
+                addError(e, true);
             }
         }
         return (Token) tokens.get(steps);
     }
 
+    /**
+     * Returns a string representation of this parser. The string will
+     * contain all the production definitions and various additional 
+     * information.
+     * 
+     * @return a detailed string representation of this parser
+     */
+    public String toString() {
+        StringBuffer  buffer = new StringBuffer();
+        
+        for (int i = 0; i < patterns.size(); i++) {
+            buffer.append(toString((ProductionPattern) patterns.get(i))); 
+            buffer.append("\n");
+        }
+        return buffer.toString();
+    }
+
+    /**
+     * Returns a string representation of a production pattern.
+     * 
+     * @param prod           the production pattern
+     * 
+     * @return a detailed string representation of the pattern
+     */
+    private String toString(ProductionPattern prod) {
+        StringBuffer  buffer = new StringBuffer();
+        StringBuffer  indent = new StringBuffer();
+        LookAheadSet  set;
+        int           i;
+
+        buffer.append(prod.getName());
+        buffer.append(" (");
+        buffer.append(prod.getId());
+        buffer.append(") ");
+        for (i = 0; i < buffer.length(); i++) {
+            indent.append(" ");
+        }
+        buffer.append("= ");
+        indent.append("| ");
+        for (i = 0; i < prod.getAlternativeCount(); i++) {
+            if (i > 0) {
+                buffer.append(indent);
+            }
+            buffer.append(toString(prod.getAlternative(i)));
+            buffer.append("\n");
+        }
+        for (i = 0; i < prod.getAlternativeCount(); i++) {
+            set = prod.getAlternative(i).getLookAhead();
+            if (set.getMaxLength() > 1) {
+                buffer.append("Using ");
+                buffer.append(set.getMaxLength());
+                buffer.append(" token look-ahead for alternative "); 
+                buffer.append(i + 1);
+                buffer.append(": ");
+                buffer.append(set.toString(tokenizer));
+                buffer.append("\n");
+            }
+        }
+        return buffer.toString();
+    }
+    
+    /**
+     * Returns a string representation of a production pattern 
+     * alternative.
+     * 
+     * @param alt            the production pattern alternative
+     * 
+     * @return a detailed string representation of the alternative
+     */
+    private String toString(ProductionPatternAlternative alt) {
+        StringBuffer  buffer = new StringBuffer();
+
+        for (int i = 0; i < alt.getElementCount(); i++) {
+            if (i > 0) {
+                buffer.append(" ");
+            }
+            buffer.append(toString(alt.getElement(i)));
+        }
+        return buffer.toString();
+    }
+
+    /**
+     * Returns a string representation of a production pattern 
+     * element.
+     * 
+     * @param elem           the production pattern element
+     * 
+     * @return a detailed string representation of the element
+     */
+    private String toString(ProductionPatternElement elem) {
+        StringBuffer  buffer = new StringBuffer();
+        int           min = elem.getMinCount();
+        int           max = elem.getMaxCount();
+
+        if (min == 0 && max == 1) {
+            buffer.append("[");
+        }
+        if (elem.isToken()) {
+            buffer.append(getTokenDescription(elem.getId()));
+        } else {
+            buffer.append(getPattern(elem.getId()).getName());
+        }
+        if (min == 0 && max == 1) {
+            buffer.append("]");
+        } else if (min == 0 && max == Integer.MAX_VALUE) {
+            buffer.append("*");
+        } else if (min == 1 && max == Integer.MAX_VALUE) {
+            buffer.append("+");
+        } else if (min != 1 || max != 1) {
+            buffer.append("{");
+            buffer.append(min);
+            buffer.append(",");
+            buffer.append(max);
+            buffer.append("}");
+        }
+        return buffer.toString();
+    }
+
     /**
      * Returns a token description for a specified token.
      * 
