@@ -44,7 +44,7 @@ import java.util.Iterator;
  * that is has to consider.
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.1
+ * @version  1.2
  */
 public class RecursiveDescentParser extends Parser {
 
@@ -186,7 +186,7 @@ public class RecursiveDescentParser extends Parser {
                 return parseAlternative(alt);
             }
         }
-        if (defaultAlt == null) {
+        if (defaultAlt == null || !isNext(defaultAlt)) {
             throwParseException(findUnion(pattern));
         }
         return parseAlternative(defaultAlt);
