@@ -40,7 +40,7 @@ package net.percederberg.grammatica.parser;
  * token patterns.
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.0
+ * @version  1.1
  */
 public class Token extends Node {
     
@@ -108,6 +108,15 @@ public class Token extends Node {
         return pattern.getId();    
     }
 
+    /**
+     * Returns the token node name.
+     * 
+     * @return the token node name
+     */
+    public String getName() {
+        return pattern.getName();
+    }
+
     /**
      * Returns the token image (i.e. the characters).
      * 
