@@ -50,14 +50,22 @@ import net.percederberg.grammatica.parser.re.RegExpException;
  * don't match any of the token patterns, a parse exception is thrown. 
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.1
+ * @version  1.2
  */
 public class Tokenizer {
     
     /**
-     * The list of all token pattern matchers.
+     * The string token matcher. This token matcher is used for all
+     * string token patterns. This matcher implements a DFA to 
+     * provide maximum performance.
      */
-    private ArrayList matchers = new ArrayList();
+    private StringTokenMatcher stringMatcher = new StringTokenMatcher();
+
+    /**
+     * The list of all regular expression token matchers. These 
+     * matchers each test matches for a single regular expression.
+     */
+    private ArrayList regexpMatchers = new ArrayList();
 
     /**
      * The input stream to read from. When this is set to null, no
@@ -114,12 +122,17 @@ public class Tokenizer {
      *         null if not present
      */
     public String getPatternDescription(int id) {
-        TokenMatcher  m;
+        TokenPattern        pattern;
+        RegExpTokenMatcher  re;
         
-        for (int i = 0; i < matchers.size(); i++) {
-            m = (TokenMatcher) matchers.get(i);
-            if (m.getPattern().getId() == id) {
-                return m.getPattern().toShortString();
+        pattern = stringMatcher.getPattern(id);
+        if (pattern != null) {
+            return pattern.toShortString();
+        }
+        for (int i = 0; i < regexpMatchers.size(); i++) {
+            re = (RegExpTokenMatcher) regexpMatchers.get(i);
+            if (re.getPattern().getId() == id) {
+                return re.getPattern().toShortString();
             }
         }
         return null;
@@ -160,11 +173,11 @@ public class Tokenizer {
 
         switch (pattern.getType()) {
         case TokenPattern.STRING_TYPE:
-            matchers.add(new StringTokenMatcher(pattern));
+            stringMatcher.addPattern(pattern);
             break;
         case TokenPattern.REGEXP_TYPE:
             try {
-                matchers.add(new RegExpTokenMatcher(pattern));
+                regexpMatchers.add(new RegExpTokenMatcher(pattern));
             } catch (RegExpException e) {
                 throw new ParserCreationException(
                     ParserCreationException.INVALID_TOKEN_ERROR,
@@ -244,9 +257,9 @@ public class Tokenizer {
 
         // Return token results
         if (m != null) {
-            str = buffer.substring(position, position + m.length());
-            token = new Token(m.getPattern(), str, line, column); 
-            position += m.length();
+            str = buffer.substring(position, position + m.getMatchedLength());
+            token = new Token(m.getMatchedPattern(), str, line, column); 
+            position += m.getMatchedLength();
             line = token.getEndLine();
             column = token.getEndColumn() + 1;
             return token;
@@ -328,17 +341,29 @@ public class Tokenizer {
      *         null if no match was found
      */
     private TokenMatcher findMatch() {
-        TokenMatcher  bestMatch = null;
-        TokenMatcher  m;
+        TokenMatcher        bestMatch = null;
+        int                 bestLength = 0;
+        RegExpTokenMatcher  re;
 
-        for (int i = 0; i < matchers.size(); i++) {
-            m = (TokenMatcher) matchers.get(i);
-            if (m.matchFrom(position)
-             && (bestMatch == null || m.length() > bestMatch.length())) {
+        // Check string matches
+        if (stringMatcher.matchFrom(position)) {
+            bestMatch = stringMatcher;
+            bestLength = bestMatch.getMatchedLength();
+        }
+        if (stringMatcher.hasReadEndOfString()) {
+            endOfBuffer = true;
+        }
 
-                bestMatch = m;
+        // Check regular expression matches
+        for (int i = 0; i < regexpMatchers.size(); i++) {
+            re = (RegExpTokenMatcher) regexpMatchers.get(i);
+            if (re.matchFrom(position)
+             && re.getMatchedLength() > bestLength) {
+
+                bestMatch = re;
+                bestLength = bestMatch.getMatchedLength();
             }
-            if (m.hasReadEndOfString()) {
+            if (re.hasReadEndOfString()) {
                 endOfBuffer = true;
             }
         }
@@ -354,12 +379,10 @@ public class Tokenizer {
      */
     public String toString() {
         StringBuffer  buffer = new StringBuffer();
-        TokenMatcher  m;
 
-        for (int i = 0; i < matchers.size(); i++) {
-            m = (TokenMatcher) matchers.get(i);
-            buffer.append(m);
-            buffer.append('\n');
+        buffer.append(stringMatcher);
+        for (int i = 0; i < regexpMatchers.size(); i++) {
+            buffer.append(regexpMatchers.get(i));
         }
         return buffer.toString();
     }
@@ -367,191 +390,51 @@ public class Tokenizer {
 
     /**
      * A token pattern matcher. This class is the base class for the
-     * two types of token matchers that exist. The token matcher 
+     * two types of token matchers that exist. The token matcher
      * checks for matches with the tokenizer buffer, and maintains the
-     * state of the last match. 
+     * state of the last match.
      */
     private abstract class TokenMatcher {
 
+        /**
+         * Returns the latest matched token pattern.
+         * 
+         * @return the latest matched token pattern, or
+         *         null if no match found
+         */
+        public abstract TokenPattern getMatchedPattern();
+        
+        /**
+         * Returns the length of the latest match.
+         * 
+         * @return the length of the latest match, or 
+         *         zero (0) if no match found
+         */
+        public abstract int getMatchedLength();
+
+        /**
+         * Checks if the end of string was encountered during the last
+         * match. 
+         * 
+         * @return true if the end of string was reached, or
+         *         false otherwise
+         */
+        public abstract boolean hasReadEndOfString();
+    }
+
+
+    /**
+     * A regular expression token pattern matcher. This class is used
+     * to match a single regular expression with the tokenizer 
+     * buffer. This class also maintains the state of the last match. 
+     */
+    private class RegExpTokenMatcher extends TokenMatcher {
+
         /**
          * The token pattern to match with.
          */
         private TokenPattern pattern;
         
-        /**
-         * Creates a new token pattern matcher.
-         * 
-         * @param pattern        the pattern to match with
-         */
-        public TokenMatcher(TokenPattern pattern) {
-            this.pattern = pattern;
-        }
-        
-        /**
-         * Returns the token pattern.
-         * 
-         * @return the token pattern
-         */
-        public TokenPattern getPattern() {
-            return pattern;
-        }
-
-        /**
-         * Returns the start position of the latest match.
-         * 
-         * @return the start position of the last match, or
-         *         zero (0) if none found
-         */
-        public abstract int start();
-
-        /**
-         * Returns the length of the latest match.
-         * 
-         * @return the length of the last match, or
-         *         zero (0) if none found
-         */        
-        public abstract int length();
-        
-        /**
-         * Checks if the end of string was encountered during the last
-         * match. 
-         * 
-         * @return true if the end of string was reached, or
-         *         false otherwise
-         */
-        public abstract boolean hasReadEndOfString(); 
-
-        /**
-         * Checks if the token pattern matches the tokenizer buffer  
-         * from the specified position. This method will also reset 
-         * all flags in this matcher.
-         * 
-         * @param pos            the starting position
-         * 
-         * @return true if a match was found, or
-         *         false otherwise
-         */
-        public abstract boolean matchFrom(int pos);
-        
-        /**
-         * Returns a string representation of this token matcher.
-         * 
-         * @return a detailed string representation of this matcher
-         */
-        public String toString() {
-            return pattern.toString() + "\n";
-        }
-    }
-    
-
-    /**
-     * A string token pattern matcher. This class is used to check 
-     * for matches with the tokenizer buffer, and maintain the state 
-     * of the last match. 
-     */
-    private class StringTokenMatcher extends TokenMatcher {
-
-        /**
-         * The string image to search for.
-         */
-        private String image;
-        
-        /**
-         * The start position of the last match. This will be a 
-         * negative value if no match was found.
-         */
-        private int start;
-
-        /**
-         * The end of string read flag.
-         */
-        private boolean endOfString;
-
-        /**
-         * Creates a new string token matcher.
-         * 
-         * @param pattern        the token pattern to use
-         */
-        public StringTokenMatcher(TokenPattern pattern) {
-            super(pattern);
-            image = pattern.getPattern();
-            reset();
-        }
-
-        /**
-         * Resets the matcher state. This will clear the results of 
-         * the last match. 
-         */
-        public void reset() {
-            start = -1;
-            endOfString = false;
-        }
-
-        /**
-         * Returns the start position of the latest match.
-         * 
-         * @return the start position of the last match, or
-         *         zero (0) if none found
-         */
-        public int start() {
-            return (start < 0) ? 0 : start;
-        }
-        
-        /**
-         * Returns the length of the latest match.
-         * 
-         * @return the length of the last match, or
-         *         zero (0) if none found
-         */        
-        public int length() {
-            return (start < 0) ? 0 : image.length();
-        }
-        
-        /**
-         * Checks if the end of string was encountered during the last
-         * match. 
-         * 
-         * @return true if the end of string was reached, or
-         *         false otherwise
-         */
-        public boolean hasReadEndOfString() {
-            return endOfString; 
-        }
-
-        /**
-         * Checks if the token pattern matches the tokenizer buffer  
-         * from the specified position. This method will also reset 
-         * all flags in this matcher.
-         * 
-         * @param pos            the starting position
-         * 
-         * @return true if a match was found, or
-         *         false otherwise
-         */
-        public boolean matchFrom(int pos) {
-            reset();
-            if (pos + image.length() > buffer.length()) {
-                endOfString = true;
-                return false;
-            }
-            for (int i = 0; i < image.length(); i++) {
-                if (image.charAt(i) != buffer.charAt(pos + i)) {
-                    return false;
-                }
-            }
-            start = pos;
-            return true;
-        }
-    }
-    
-
-    /**
-     * A regular expression token pattern matcher. This class is used
-     * to check for matches with the tokenizer buffer, and maintain 
-     * the state of the last match. 
-     */
-    private class RegExpTokenMatcher extends TokenMatcher {
-
         /**
          * The regular expression to use.
          */
@@ -573,9 +456,18 @@ public class Tokenizer {
         public RegExpTokenMatcher(TokenPattern pattern)
             throws RegExpException {
 
-            super(pattern);
-            regExp = new RegExp(pattern.getPattern());
-            matcher = regExp.matcher(buffer);
+            this.pattern = pattern;
+            this.regExp = new RegExp(pattern.getPattern());
+            this.matcher = regExp.matcher(buffer);
+        }
+
+        /**
+         * Returns the token pattern.
+         * 
+         * @return the token pattern
+         */
+        public TokenPattern getPattern() {
+            return pattern;
         }
 
         /**
@@ -592,13 +484,27 @@ public class Tokenizer {
             }
         }
         
+        /**
+         * Returns the latest matched token pattern.
+         * 
+         * @return the latest matched token pattern, or
+         *         null if no match found
+         */
+        public TokenPattern getMatchedPattern() {
+            if (matcher.length() <= 0) {
+                return null;
+            } else {
+                return pattern;
+            }
+        }
+        
         /**
          * Returns the length of the latest match.
          * 
-         * @return the length of the last match, or
-         *         zero (0) if none found
-         */        
-        public int length() {
+         * @return the length of the latest match, or 
+         *         zero (0) if no match found
+         */
+        public int getMatchedLength() {
             return matcher.length();
         }
         
@@ -633,7 +539,313 @@ public class Tokenizer {
          * @return a detailed string representation of this matcher
          */
         public String toString() {
-            return super.toString() + regExp.toString();
+            return pattern.toString() + "\n" + 
+                   regExp.toString() + "\n";
+        }
+    }
+
+
+    /**
+     * A string token pattern matcher. This class is used to match a 
+     * set of strings with the tokenizer buffer. This class 
+     * internally uses a DFA for maximum performance. It also 
+     * maintains the state of the last match. 
+     */
+    private class StringTokenMatcher extends TokenMatcher {
+
+        /**
+         * The list of string token patterns. 
+         */
+        private ArrayList patterns = new ArrayList();
+
+        /**
+         * The finite automaton to use for matching.
+         */
+        private Automaton start = new Automaton();
+
+        /**
+         * The last token pattern match found.
+         */
+        private TokenPattern match = null;
+
+        /**
+         * The end of string read flag.
+         */
+        private boolean endOfString = false;
+
+        /**
+         * Creates a new string token matcher.
+         */
+        public StringTokenMatcher() {
+        }
+
+        /**
+         * Resets the matcher state. This will clear the results of 
+         * the last match. 
+         */
+        public void reset() {
+            match = null;
+            endOfString = false;
+        }
+
+        /**
+         * Returns the latest matched token pattern.
+         * 
+         * @return the latest matched token pattern, or
+         *         null if no match found
+         */
+        public TokenPattern getMatchedPattern() {
+            return match;
+        }
+        
+        /**
+         * Returns the length of the latest match.
+         * 
+         * @return the length of the latest match, or 
+         *         zero (0) if no match found
+         */
+        public int getMatchedLength() {
+            if (match == null) {
+                return 0;
+            } else {
+                return match.getPattern().length();
+            }
+        }
+
+        /**
+         * Checks if the end of string was encountered during the last
+         * match. 
+         * 
+         * @return true if the end of string was reached, or
+         *         false otherwise
+         */
+        public boolean hasReadEndOfString() {
+            return endOfString; 
+        }
+
+        /**
+         * Sets the end of string encountered flag.
+         */
+        public void setReadEndOfString() {
+            endOfString = true;
+        }
+
+        /**
+         * Returns the token pattern with the specified id. Only 
+         * token patterns handled by this matcher can be returned. 
+         * 
+         * @param id         the token pattern id
+         * 
+         * @return the token pattern found, or
+         *         null if not found
+         */
+        public TokenPattern getPattern(int id) {
+            TokenPattern  pattern;
+
+            for (int i = 0; i < patterns.size(); i++) {
+                pattern = (TokenPattern) patterns.get(i);
+                if (pattern.getId() == id) {
+                    return pattern;
+                }
+            }
+            return null;
+        }
+
+        /**
+         * Adds a string token pattern to this matcher. 
+         * 
+         * @param pattern        the pattern to add
+         */
+        public void addPattern(TokenPattern pattern) {
+            patterns.add(pattern);
+            start.addMatch(pattern.getPattern(), pattern);
+        }
+
+        /**
+         * Checks if the token pattern matches the tokenizer buffer  
+         * from the specified position. This method will also reset 
+         * all flags in this matcher.
+         * 
+         * @param pos            the starting position
+         * 
+         * @return true if a match was found, or
+         *         false otherwise
+         */
+        public boolean matchFrom(int pos) {
+            reset();
+            match = (TokenPattern) start.matchFrom(this, pos);
+            return match != null;
+        }
+        
+        /**
+         * Returns a string representation of this matcher. This will
+         * contain all the token patterns.
+         * 
+         * @return a detailed string representation of this matcher 
+         */
+        public String toString() {
+            StringBuffer  buffer = new StringBuffer();
+            
+            for (int i = 0; i < patterns.size(); i++) {
+                buffer.append(patterns.get(i));
+                buffer.append("\n\n");
+            }
+            return buffer.toString();
+        }
+    }
+    
+
+    /**
+     * A deterministic finite automaton. This is a simple automaton
+     * for character sequences. It cannot handle character set state
+     * transitions, but only supports single character transitions.
+     */
+    private class Automaton {
+
+        /**
+         * The state value.
+         */
+        private Object value = null;
+        
+        /**
+         * The automaton state transition tree. Each transition from
+         * this state to another state is added to this tree with the
+         * corresponding character. 
+         */
+        private AutomatonTree tree = new AutomatonTree();
+        
+        /**
+         * Creates a new empty automaton. 
+         */
+        public Automaton() {
+        }
+
+        /**
+         * Adds a string match to this automaton. New states and 
+         * transitions will be added to extend this automaton to 
+         * support the specified string.
+         * 
+         * @param str            the string to match
+         * @param value          the match value
+         */
+        public void addMatch(String str, Object value) {
+            Automaton  state;
+
+            if (str.equals("")) {
+                this.value = value;
+            } else {
+                state = tree.find(str.charAt(0));
+                if (state == null) {
+                    state = new Automaton();
+                    state.addMatch(str.substring(1), value);
+                    tree.add(str.charAt(0), state);
+                } else {
+                    state.addMatch(str.substring(1), value);
+                }
+            }
+        }
+
+        /**
+         * Checks if the automaton matches the tokenizer buffer from 
+         * the specified position. This method will set the end of
+         * buffer flag in the specified token matcher if the end of
+         * buffer is reached.
+         *
+         * @param m              the string token matcher 
+         * @param pos            the starting position
+         * 
+         * @return the match value, or
+         *         null if no match is found
+         */
+        public Object matchFrom(StringTokenMatcher m, int pos) {
+            Object     result = null;
+            Automaton  state;
+
+            if (pos >= buffer.length()) {
+                m.setReadEndOfString();
+            } else if (tree != null) {
+                state = tree.find(buffer.charAt(pos));
+                if (state != null) {
+                    result = state.matchFrom(m, pos + 1);
+                }
+            }
+            return (result == null) ? value : result;
+        }
+    }
+    
+    
+    /**
+     * An automaton state transition tree. This class contains a 
+     * binary search tree for the automaton transitions from one 
+     * state to another. All transitions are linked to a single 
+     * character.
+     */
+    private class AutomatonTree {
+
+        /**
+         * The transition character. If this value is set to the zero 
+         * ('\0') character, this tree is empty.
+         */
+        private char value = '\0';
+        
+        /**
+         * The transition state.
+         */
+        private Automaton state = null;
+        
+        /**
+         * The left subtree.
+         */
+        private AutomatonTree left = null;
+        
+        /**
+         * The right subtree.
+         */
+        private AutomatonTree right = null;
+        
+        /**
+         * Creates a new empty automaton transition tree.
+         */
+        public AutomatonTree() {
+        }
+        
+        /**
+         * Finds an automaton state from the specified transition 
+         * character. This method searches this transition tree for
+         * a matching transition.
+         * 
+         * @param c              the character to search for
+         * 
+         * @return the automaton state found, or
+         *         null if no transition exists
+         */
+        public Automaton find(char c) {
+            if (value == '\0' || value == c) {
+                return state;
+            } else if (value > c) {
+                return left.find(c); 
+            } else {
+                return right.find(c);
+            }
+        }
+        
+        /**
+         * Adds a transition to this tree. 
+         * 
+         * @param c              the character to transition for
+         * @param state          the state to transition to
+         */
+        public void add(char c, Automaton state) {
+            if (value == '\0') {
+                this.value = c;
+                this.state = state;
+                this.left = new AutomatonTree();
+                this.right = new AutomatonTree();
+            } else if (value > c) {
+                left.add(c, state);
+            } else {
+                right.add(c, state);
+            }
         }
     }
 }
