@@ -402,12 +402,73 @@ public class TestRegExp extends TestCase {
         failMatchRegExp("a{2,3}+a", "aaa");
     }
 
+    /**
+     * Tests the quantifier backtracking for stack overflows.
+     * (Bug #3632)
+     */
+    public void testQuantifierStackOverflow() {
+        StringBuffer  buffer = new StringBuffer();
+        String        str;
+        
+        for (int i = 0; i < 4096; i++) {
+            buffer.append("a");
+        }
+        str = buffer.toString();
+        matchRegExp("a*" + str, str);
+        failMatchRegExp("a*a" + str, str);
+        matchRegExp("a*?b", str + "b");
+        failMatchRegExp("a*?b", str);
+        matchRegExp("a*+", str);
+        failMatchRegExp("a*+a", str);
+    }
+
+    /**
+     * Tests the quantifier backtracking state memory. This tests 
+     * changes the character buffer used by the matcher object and
+     * tests if the quantifier state is properly reset. (Bug #3653)
+     */
+    public void testCharBufferAppend() {
+        Matcher     m;
+        CharBuffer  buffer;
+        
+        buffer = new CharBuffer();
+        m = createRegExp("a*aa").matcher(buffer);
+        buffer.append("a");
+        if (m.matchFromBeginning()) {
+            fail("found invalid match '" + m.toString() + 
+                 "' to regexp 'a*aa' in input '" + buffer + "'");
+        }
+        buffer.append("aaa");
+        if (!m.matchFromBeginning()) {
+            fail("couldn't match '" + buffer + "' to regexp 'a*aa'");
+        } else if (!buffer.toString().equals(m.toString())) {
+            fail("incorrect match for 'a*aa', found: '" + 
+                 m.toString() + "', expected: '" + buffer + "'");
+        }
+
+        buffer = new CharBuffer();
+        m = createRegExp("a*?b").matcher(buffer);
+        buffer.append("aaa");
+        if (m.matchFromBeginning()) {
+            fail("found invalid match '" + m.toString() + 
+                 "' to regexp 'a*?b' in input '" + buffer + "'");
+        }
+        buffer.append("aab");
+        if (!m.matchFromBeginning()) {
+            fail("couldn't match '" + buffer + "' to regexp 'a*?b'");
+        } else if (!buffer.toString().equals(m.toString())) {
+            fail("incorrect match for 'a*?b', found: '" + 
+                 m.toString() + "', expected: '" + buffer + "'");
+        }
+    }
+
     /**
      * Tests matching of various logical operators. 
      */
     public void testLogicalOperators() {
-        matchRegExp("a|b", "a");
-        matchRegExp("a|b", "b");
+        matchRegExp("a|ab|b", "a");
+        matchRegExp("a|ab|b", "b");
+        matchRegExp("a|ab|b", "ab");
         matchRegExp("(ab)", "ab");
         matchRegExp("(a)(b)", "ab");
     }
